@@ -1,26 +1,78 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of this
-#repository contains the full copyright notices and license terms.
-
-from trytond.model import ModelSQL, ModelView, Workflow, fields
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
+from trytond import backend
+from trytond.model import ModelSQL, ModelView, Workflow, fields, tree
 from trytond.pool import Pool, PoolMeta
 from trytond.pyson import Eval
 from trytond.report import Report
 
-__all__ = ['Configuration', 'Package', 'Type', 'Move',
+__all__ = ['Configuration', 'ConfigurationSequence', 'Package', 'Type', 'Move',
     'ShipmentOut', 'ShipmentInReturn', 'PackageLabel']
 
 
-class Configuration:
-    __metaclass__ = PoolMeta
+class Configuration(metaclass=PoolMeta):
     __name__ = 'stock.configuration'
-    package_sequence = fields.Property(fields.Many2One('ir.sequence',
-            'Package Sequence', domain=[
+    package_sequence = fields.MultiValue(fields.Many2One(
+            'ir.sequence', "Package Sequence", required=True,
+            domain=[
                 ('company', 'in', [Eval('context', {}).get('company'), None]),
                 ('code', '=', 'stock.package'),
-                ], required=True))
+                ]))
+
+    @classmethod
+    def multivalue_model(cls, field):
+        pool = Pool()
+        if field == 'package_sequence':
+            return pool.get('stock.configuration.sequence')
+        return super(Configuration, cls).multivalue_model(field)
+
+    @classmethod
+    def default_package_sequence(cls, **pattern):
+        return cls.multivalue_model(
+            'package_sequence').default_package_sequence()
 
 
-class Package(ModelSQL, ModelView):
+class ConfigurationSequence(metaclass=PoolMeta):
+    __name__ = 'stock.configuration.sequence'
+    package_sequence = fields.Many2One(
+        'ir.sequence', "Package Sequence", required=True,
+        domain=[
+            ('company', 'in', [Eval('company', -1), None]),
+            ('code', '=', 'stock.package'),
+            ],
+        depends=['company'])
+
+    @classmethod
+    def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
+        exist = TableHandler.table_exist(cls._table)
+        if exist:
+            table = cls.__table_handler__(module_name)
+            exist &= table.column_exist('package_sequence')
+
+        super(ConfigurationSequence, cls).__register__(module_name)
+
+        if not exist:
+            cls._migrate_property([], [], [])
+
+    @classmethod
+    def _migrate_property(cls, field_names, value_names, fields):
+        field_names.append('package_sequence')
+        value_names.append('package_sequence')
+        super(ConfigurationSequence, cls)._migrate_property(
+            field_names, value_names, fields)
+
+    @classmethod
+    def default_package_sequence(cls):
+        pool = Pool()
+        ModelData = pool.get('ir.model.data')
+        try:
+            return ModelData.get_id('stock_package', 'sequence_package')
+        except KeyError:
+            return None
+
+
+class Package(tree(), ModelSQL, ModelView):
     'Stock Package'
     __name__ = 'stock.package'
     _rec_name = 'code'
@@ -32,6 +84,7 @@ class Package(ModelSQL, ModelView):
         domain=[
             ('shipment', '=', Eval('shipment')),
             ('to_location.type', 'in', ['customer', 'supplier']),
+            ('state', '!=', 'cancel'),
             ],
         add_remove=[
             ('package', '=', None),
@@ -74,11 +127,6 @@ class Package(ModelSQL, ModelView):
             values['code'] = Sequence.get_id(config.package_sequence)
         return super(Package, cls).create(vlist)
 
-    @classmethod
-    def validate(cls, packages):
-        super(Package, cls).validate(packages)
-        cls.check_recursion(packages)
-
 
 class Type(ModelSQL, ModelView):
     'Stock Package Type'
@@ -86,13 +134,12 @@ class Type(ModelSQL, ModelView):
     name = fields.Char('Name', required=True)
 
 
-class Move:
-    __metaclass__ = PoolMeta
+class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
     package = fields.Many2One('stock.package', 'Package', select=True)
 
 
-class PackageMixin:
+class PackageMixin(object):
     packages = fields.One2Many('stock.package', 'shipment', 'Packages',
         states={
             'readonly': Eval('state').in_(['done', 'cancel']),
@@ -115,9 +162,21 @@ class PackageMixin:
                 'packages': value,
                 })
 
+    @classmethod
+    def check_packages(cls, shipments):
+        for shipment in shipments:
+            if not shipment.packages:
+                continue
+            length = sum(len(p.moves) for p in shipment.packages)
+            if len(list(shipment.packages_moves)) != length:
+                cls.raise_user_error('package_mismatch', shipment.rec_name)
 
-class ShipmentOut(object, PackageMixin):
-    __metaclass__ = PoolMeta
+    @property
+    def packages_moves(self):
+        raise NotImplementedError
+
+
+class ShipmentOut(PackageMixin, object, metaclass=PoolMeta):
     __name__ = 'stock.shipment.out'
 
     @classmethod
@@ -128,28 +187,33 @@ class ShipmentOut(object, PackageMixin):
                     'Customer Shipment "%s" are packaged.'),
                 })
 
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('packed')
+    def pack(cls, shipments):
+        super(ShipmentOut, cls).pack(shipments)
+        cls.check_packages(shipments)
+
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
     def done(cls, shipments):
         super(ShipmentOut, cls).done(shipments)
-        for shipment in shipments:
-            if not shipment.packages:
-                continue
-            if (len(shipment.outgoing_moves)
-                    != sum(len(p.moves) for p in shipment.packages)):
-                cls.raise_user_error('package_mismatch', shipment.rec_name)
+        cls.check_packages(shipments)
+
+    @property
+    def packages_moves(self):
+        return (m for m in self.outgoing_moves if m.state != 'cancel')
 
 
-class ShipmentInReturn(object, PackageMixin):
-    __metaclass__ = PoolMeta
+class ShipmentInReturn(PackageMixin, object, metaclass=PoolMeta):
     __name__ = 'stock.shipment.in.return'
 
     @classmethod
     def __setup__(cls):
         super(ShipmentInReturn, cls).__setup__()
         cls._error_messages.update({
-                'package_mismatch': ('Not all Outgoing Moves of '
+                'package_mismatch': ('Not all Moves of '
                     'Supplier Return Shipment "%s" are packaged.'),
                 })
 
@@ -158,12 +222,11 @@ class ShipmentInReturn(object, PackageMixin):
     @Workflow.transition('done')
     def done(cls, shipments):
         super(ShipmentInReturn, cls).done(shipments)
-        for shipment in shipments:
-            if not shipment.packages:
-                continue
-            if (len(shipment.moves)
-                    != sum(len(p.moves) for p in shipment.packages)):
-                cls.raise_user_error('package_mismatch', shipment.rec_name)
+        cls.check_packages(shipments)
+
+    @property
+    def packages_moves(self):
+        return (m for m in self.moves if m.state != 'cancel')
 
 
 class PackageLabel(Report):
